/* os/linux-gnu/affinity.c */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

#define _POSIX_C_SOURCE 200112L
#define _ISOC99_SOURCE
#define _XOPEN_SOURCE 600
#define _GNU_SOURCE

#include <limits.h>
#include <sched.h>
#include <pthread.h>
#include <errno.h>
#include <unistd.h>

#include <affinity.h>

/* taken from sched_setaffinity(2) */
#define OUR_MAX_CPU_COUNT (1024 << 8)

static int our_cpu_count = sizeof(cpu_set_t);

/* Sets the affinity of the given thread to the given CPU if newcpu is
 * not INT_MAX, and returns the previous affinity of the given thread in the
 * buffer pointed to by oldcpu if oldcpu is not NULL.
 *
 * This function is intended to be platform-agnostic, in the absence of a POSIX
 * function to do this.
 *
 * On success, this function returns 0.  On error, error code will be placed
 * into errno and this function will return -1.
 *
 * This function will fail if no memory for CPU sets is available, the number of
 * CPUs in the system cannot be reliably determined, or the affinity cannot be
 * set for some reason. */
int
_exs_set_thread_affinity(pthread_t threadid, int newcpu, int *oldcpu)
{
	int ncpus;
	int i;
	int errcode;

	ncpus = _exs_cpu_count();
	if (ncpus < 0) {
		return -1;
	}

	if (oldcpu) {
		cpu_set_t *oldset;
		size_t oldsetsize;

		oldsetsize = CPU_ALLOC_SIZE(ncpus);
		oldset = CPU_ALLOC(ncpus);
		if (oldset == NULL) {
			errno = ENOMEM;
			return -1;
		}

		if ((errcode = pthread_getaffinity_np(threadid, oldsetsize,
								oldset)) != 0) {
			errno = errcode;
			return -1;
		}

		*oldcpu = INT_MAX;
		for (i = 0; i < ncpus; ++i) {
			if (CPU_ISSET_S(i, oldsetsize, oldset)) {
				if (*oldcpu == INT_MAX) {
					*oldcpu = i;
				} else {
					*oldcpu = INT_MAX;
					break;
				}
			}
		}

		CPU_FREE(oldset);
	}

	if (newcpu != INT_MAX) {
		cpu_set_t *newset;
		size_t newsetsize;

		newsetsize = CPU_ALLOC_SIZE(ncpus);
		newset = CPU_ALLOC(ncpus);
		if (newset == NULL) {
			errno = ENOMEM;
			return -1;
		}

		CPU_ZERO_S(newsetsize, newset);
		CPU_SET_S(newcpu, newsetsize, newset);

		if ((errcode = pthread_setaffinity_np(threadid, newsetsize,
								newset)) != 0) {
			errno = errcode;
			return -1;
		}

		CPU_FREE(newset);
	}

	return 0;
}

/* Returns the number of CPUs that should be used for making CPU sets.  On
 * failure, returns -1 and sets errno to ERANGE. */
int
_exs_cpu_count(void)
{
	int next_guess;
	int guess;
	size_t bytes;
	cpu_set_t *set;
	int retval;

	next_guess = our_cpu_count;
	do {
		guess = next_guess;
		next_guess = guess * 2;
		set = CPU_ALLOC(guess);
		bytes = CPU_ALLOC_SIZE(guess);
		CPU_ZERO_S(bytes, set);
		retval = sched_getaffinity(0, bytes, set);
		CPU_FREE(set);
	} while (retval == -1 && errno == EINVAL && guess < OUR_MAX_CPU_COUNT);

	if (retval == -1 && errno == EINVAL) {
		/* Too many CPUs */
		errno = ERANGE;
		return -1;
	}

	return guess;
}
